package nq.net.assets {

    import flash.events.Event;
    import flash.events.EventDispatcher;

    import nq.net.assets.loaders.IAssetLoader;
    import nq.net.assets.loaders.IAssetLoaderFactory;
    import nq.net.assets.storage.AssetBundle;
    import nq.net.assets.storage.IAssetBundle;
    import nq.utils.task.TaskQueue;
    import nq.utils.task.TaskQueueEvent;

    /**
     * Событие удачной загрузки ресурса
     */
    [Event(name = "assetLoaded", type = "nq.net.assets.AssetManagerEvent")]

    /**
     * Событие неудачи при загрузке ресурса
     */
    [Event(name = "assetFailed", type = "nq.net.assets.AssetManagerEvent")]

    /**
     * Многопоточный загрузчик ресурсов, складывающий результаты
     * загрузки в указанное хранилище. Загрузчик является оберткой
     * вокруг хранилища ресурсов
     */
    public final class AssetManager extends EventDispatcher implements IAssetManager {

        private var _factory:IAssetLoaderFactory;
        private var _taskQueue:TaskQueue;
        private var _tasks:Object; // url -> task
        private var _storage:IAssetBundle;
        private var _retryCount:uint;
        private var _beingLoaded:Object;

        /**
         * @param loaderFactory Фабрика для генерации загрузчиков под определенный тип контента
         * @param storage       Хранилище загруженных ресурсов (если не указано, то создается новое хранилище)
         * @param retryCount    Максимальное число попыток загрузки ресурса
         * @param maxThreads    Максимальное число потоков загрузки
         */
        public function AssetManager(loaderFactory:IAssetLoaderFactory, storage:IAssetBundle = null, retryCount:uint = 1, maxThreads:uint = 5) {

            super();

            if (!loaderFactory) throw new Error("Фабрика загрузчиков должна быть не нулевой!");
            _factory = loaderFactory;

            _tasks = {};
            _storage = storage ? storage : new AssetBundle();
            _retryCount = retryCount;
            _taskQueue = new TaskQueue(maxThreads, true);

            _beingLoaded = new Object();

            _taskQueue.addEventListener(TaskQueueEvent.TASK_COMPLETE, onTaskComplete);
            _taskQueue.addEventListener(TaskQueueEvent.TASK_FAIL, onTaskFail);

        }

        private function onTaskFail(event:TaskQueueEvent):void {

            var url:String = (event.task as LoadAssetTask).url;
            var numRetries:int = int(_beingLoaded[url]);

            if (numRetries < _retryCount) {

                _beingLoaded[url] = numRetries + 1;

                _taskQueue.addTask(event.task);
                if (!_taskQueue.isRunning) _taskQueue.start();

            }else {

                delete _beingLoaded[url];
                super.dispatchEvent(new AssetManagerEvent(AssetManagerEvent.ASSET_FAILED, url));

            }

        }

        private function onTaskComplete(event:TaskQueueEvent):void {

            var url:String = (event.task as LoadAssetTask).url;

            delete _beingLoaded[url];

            super.dispatchEvent(new AssetManagerEvent(AssetManagerEvent.ASSET_LOADED, url));
        }


        /* INTERFACE nq.net.assets.IAssetManager */

        /**
         * Добавить ресурс к списку загружаемых
         *
         * @param url       URL-адрес ресурса для загрузки
         * @param mimeType  MIME-тип ресурса для загрузки
         * @param priority  Приоритет загрузки ресурса
         */
        public function load(url:String, mimeType:String = "application/octet-stream" /* nq.net.MIMEType.OCTET_STREAM */, priority:int = 0):void {

            if (_tasks[url])
            {
                var existingTask:LoadAssetTask = _tasks[url];
                if (existingTask.mimeType != mimeType)
                {
                    throw new Error("Поытка загрузить один url два раза с разными mime type");
                }

                if (existingTask.priority < priority && !_taskQueue.isTaskRunning(existingTask))
                {
                    _taskQueue.removeTask(existingTask);
                }
                else
                {
                    return;
                }
            }

            _beingLoaded[url] = 0;

            var loader:IAssetLoader = _factory.getLoader(url, mimeType);
            var task:LoadAssetTask = new LoadAssetTask(_storage, loader, url, priority, mimeType);

            _tasks[url] = task;

            _taskQueue.addTask(task);
            if (!_taskQueue.isRunning) _taskQueue.start();

        }

        /**
         * Добавить ассет в пакет ресурсов
         *
         * @param asset Добавляемый ассет
         * @param uri   Уникальное имя ассета в этом пакете
         */
        public function addAsset(asset:Object, uri:String):void {

            _storage.addAsset(asset,  uri);
            super.dispatchEvent(new Event(Event.ADDED));
        }

        /**
         * Имеется ли ассет с таким именем в пакете
         *
         * @param uri   Уникальное имя ассета в этом пакете
         */
        public function hasAsset(uri:String):Boolean {
            return _storage.hasAsset(uri);
        }

        /**
         * Удалить ассет по имени
         *
         * @param uri   Уникальное имя ассета в этом пакете
         */
        public function removeAsset(uri:String):void {

            _storage.removeAsset(uri);
            super.dispatchEvent(new Event(Event.REMOVED));
        }

        /**
         * Получить ассет по имени
         *
         * @param uri   Уникальное имя ассета в этом пакете
         */
        public function getAsset(uri:String):Object {
            return _storage.getAsset(uri);
        }

    }

}